"""
Interactions with the user through the cli
"""
from typing import Any, Callable, List, Tuple

RolePrincipalPair = Tuple


def user_choice(question: str,
                options: List[Any],
                renderer: Callable[[Any], str] = lambda x: x,
                saved_choice: str = None):
    """
    Prompt a user with a question and a specific set of possible responses
    :param question: Specifying context for the user to select an option
    :param options: A list of options for the user to select from
    """
    if saved_choice:
        for option in options:
            if renderer(option) == saved_choice:
                return option
        print("Ignoring invalid saved choice '{}'".format(saved_choice))

    if len(options) == 1:
        return options[0]

    print(question)
    if len(options) == 0:
        raise Exception("No options found")
    option_list = []
    for i, option in enumerate(options):
        option_list.append("[{}] {}".format(i + 1, renderer(option)))
    selection = None
    while selection is None:
        print("\n".join(option_list))
        choice = input("? ")
        try:
            val = int(choice) - 1
            if val in range(0, len(options)):
                selection = options[val]
            else:
                print("Invalid option")
        except ValueError:
            print("Invalid option")
    return selection


def user_role_prompt(all_roles: List[RolePrincipalPair],
                     saved_choice: str = None) -> RolePrincipalPair:
    """
    Prompt a user with a list of AWS IAM roles to choose from. If only 1 role
    is available, return that.
    """
    return user_choice(
        "Pick a role:",
        all_roles,
        renderer=lambda role: role[0],
        saved_choice=saved_choice,
    )



"""OneLogin/AWS Business logic"""

from typing import Optional

import configparser
import xml.etree.ElementTree as ElementTree

import base64
import boto3
import os
import re

from requests import get

from onelogin.api.client import OneLoginClient

from onelogin_aws_login.configuration import Section
from onelogin_aws_login.credentials import MFACredentials, UserCredentials
from onelogin_aws_login.userquery import user_role_prompt

CONFIG_FILENAME = ".onelogin-aws.config"
DEFAULT_CONFIG_PATH = os.path.join(os.path.expanduser("~"), CONFIG_FILENAME)


class OneloginAWS(object):
    """
    Handles the authentication between OneLogin SAML Assertion and the AWS
    identity federation
    """

    def __init__(self, config: Section):
        self.sts_client = boto3.client("sts")
        self.config = config
        self.saml = None
        self.all_roles = None
        self.role_arn = None
        self.credentials = None
        self.duration_seconds = int(config['duration_seconds'])
        self.user_credentials = UserCredentials(config)
        self.mfa = MFACredentials(config)

        base_uri_parts = self.config['base_uri'].split('.')
        self.ol_client = OneLoginClient(
            self.config['client_id'],
            self.config['client_secret'],
            base_uri_parts[1],
        )

    def check_for_errors(self, response):
        """
        Throw exceptions that the OneLogin API should have thrown
        """

        if self.ol_client.error is None:
            return response

        raise Exception("Onelogin Error: '{error}' '{desc}'".format(
            error=self.ol_client.error,
            desc=self.ol_client.error_description
        ))

    def get_saml_assertion(self):
        """
        Retrieve users credentials and get the SAML assertion from Onelogin,
        based on the users choice of AWS account to log into
        """

        self.user_credentials.load_credentials()

        saml_resp = self.check_for_errors(
            self.ol_client.get_saml_assertion(
                username_or_email=self.user_credentials.username,
                password=self.user_credentials.password,
                app_id=self.config['aws_app_id'],
                subdomain=self.config['subdomain'],
                ip_address=self.get_ip_address(),
            )
        )

        if saml_resp.mfa:
            if not self.mfa.ready():
                self.mfa.select_device(saml_resp.mfa.devices)
                if not self.mfa.has_otp:
                    self.mfa.prompt_token()

            saml_resp = self.check_for_errors(
                self.ol_client.get_saml_assertion_verifying(
                    self.config['aws_app_id'],
                    self.mfa.device.id,
                    saml_resp.mfa.state_token,
                    self.mfa.otp
                )
            )

        self.saml = saml_resp

    def get_ip_address(self) -> Optional[str]:
        """
        Get the client IP address.
        Uses either the `ip_address` in config,
        or if `auto_determine_ip_address` is specified in config,
        the ipify service is used to dynamically lookup the IP address.
        """

        # if ip address has been hard coded in config file, use that
        ip_address = self.config.get('ip_address')
        if ip_address is not None:
            return ip_address

        # if auto determine is enabled, use ipify to lookup the ip
        if self.config.auto_determine_ip_address:
            ip_address = get('https://api.ipify.org').text
            return ip_address

    def get_arns(self):
        """Extract the IAM Role ARNs from the SAML Assertion"""

        if not self.saml:
            self.get_saml_assertion()
        # Parse the returned assertion and extract the authorized roles
        aws_roles = []
        root = ElementTree.fromstring(
            base64.b64decode(self.saml.saml_response))

        namespace = "{urn:oasis:names:tc:SAML:2.0:assertion}"
        role_name = "https://aws.amazon.com/SAML/Attributes/Role"
        for attr in root.iter(namespace + "Attribute"):
            if attr.get("Name") == role_name:
                for val in attr.iter(namespace + "AttributeValue"):
                    aws_roles.append(val.text)

        # Note the format of the attribute value should be role_arn,
        # principal_arn but lots of blogs list it as principal_arn,role_arn so
        # let's reverse them if needed
        aws_roles = [role.split(",") for role in aws_roles]
        aws_roles = [(role, principal) for role, principal in aws_roles]
        self.all_roles = aws_roles

    def get_role(self):
        """
        Prompt the user to choose a Role ARN if more than one is available
        """

        if not self.all_roles:
            self.get_arns()

        if not self.all_roles:
            raise Exception("No roles found")

        # If I have more than one role, ask the user which one they want,
        # otherwise just proceed

        self.role_arn, self.principal_arn = user_role_prompt(
            self.all_roles,
            saved_choice=self.config.get("role_arn"),
        )

    def assume_role(self):
        """Perform an AWS SAML role assumption"""

        if not self.role_arn:
            self.get_role()
        if self.config['region']:
            self.sts_client = boto3.client(
                "sts",
                region_name=self.config["region"])
        res = self.sts_client.assume_role_with_saml(
            RoleArn=self.role_arn,
            PrincipalArn=self.principal_arn,
            SAMLAssertion=self.saml.saml_response,
            DurationSeconds=self.duration_seconds
        )

        self.credentials = res

    def save_credentials(self):
        """Save the AWS Federation credentials to disk"""

        if not self.credentials:
            self.assume_role()

        creds = self.credentials["Credentials"]

        cred_file = self._initialize_credentials()

        cred_config = configparser.ConfigParser()
        cred_config.read(cred_file)

        # Update with new credentials
        name = self.credentials["AssumedRoleUser"]["Arn"]
        m = re.search(r'(arn\:aws([\w-]*)\:sts\:\:)(.*)', name)

        if m is not None:
            name = m.group(3)
        name = name.replace(":assumed-role", "")
        if "profile" in self.config:
            name = self.config["profile"]

        # Initialize the profile block if it is undefined
        if name not in cred_config:
            cred_config[name] = {}

        # Set each value specifically instead of overwriting the entire
        # profile block in case they have other parameters defined
        cred_config[name]['aws_access_key_id'] = creds["AccessKeyId"]
        cred_config[name]['aws_secret_access_key'] = creds["SecretAccessKey"]
        cred_config[name]['aws_session_token'] = creds["SessionToken"]

        # Set region for this profile if passed in via configuration
        if self.config['region']:
            cred_config[name]['region'] = self.config['region']

        with open(cred_file, "w") as cred_config_file:
            cred_config.write(cred_config_file)

        print("Credentials cached in '{}'".format(cred_file))
        print("Expires at {}".format(creds["Expiration"]))
        print("Use aws cli with --profile " + name)

        # Reset state in the case of another transaction
        self.credentials = None

    def _initialize_credentials(self):
        cred_file = os.environ.get('AWS_SHARED_CREDENTIALS_FILE', None)

        if cred_file is None:
            cred_file = os.path.expanduser("~/.aws/credentials")
            cred_dir = os.path.expanduser("~/.aws/")
            if not os.path.exists(cred_dir):
                os.makedirs(cred_dir)

        return cred_file



"""
Handles the saving and loading of username and password in a secure
manner
"""
import getpass
from typing import List

import keyring
from onelogin.api.models.device import Device

from onelogin_aws_login.configuration import Section
from onelogin_aws_login.userquery import user_choice


class MFACredentials(object):
    """
    Class to encapsulate the handling and storage of MFA devices, and
    retrieving of OTP's
    """

    def __init__(self, config: Section):
        self._devices = []
        self.device = None
        self._otp = None
        self._config = config

        self.reset()

    @property
    def has_device(self) -> bool:
        """True if the MFA has an MFA device selected waiting to be used"""
        return self.device is not None

    @property
    def has_otp(self) -> bool:
        """True if the MFA has an OTP waiting to be used"""
        return self._otp is not None

    @property
    def otp(self) -> str:
        """
        Return the OTP for the MFA and reset the OTP.
        OTP's can only be used once, so it will be reset after.
        """
        result = self._otp
        self._otp = None
        return result

    def ready(self):
        """If the MFA is ready to be used"""
        return self.has_otp and self.has_device

    def reset(self):
        """Remove all state from this class"""

        self._devices = []
        self.device = None

        self._otp = None

    def select_device(self, devices: List[Device]):
        """Given a list of MFA devices, select one for use"""

        self._devices = devices

        self.device = user_choice(
            'Pick an OTP Device:',
            self._devices,
            renderer=lambda d: d.type,
            saved_choice=self._config.get("otp_device"),
        )

    def prompt_token(self):
        """Ask the user for an OTP token"""
        self._otp = input("{device} Token: ".format(device=self.device.type))


class UserCredentials(object):
    """
    Class to encapsulate the handling of storing and retrieving user password
    in OS-Independent system keychain.
    """
    SERVICE_NAME = "onelogin-aws-cli"

    def __init__(self, config: Section):
        self.username = config.get('username')
        self.configuration = config

        # This is `None`, as the password should be be emitted from this class
        # and should never be loaded from any other source outside this class
        self.password = None

    @property
    def has_password(self) -> bool:
        """
        True if the class has a password.

        :return: Whether we have set a password or not, yet
        """
        return (self.password is not None) and \
               (self.password != "")

    def load_credentials(self):
        """Load the username and password"""

        self.load_username()
        self.load_password()

    def load_username(self):
        """
        Either load the username from configfile or prompt the user to supply
        one interactively
        """

        if not self.username:
            # Try the configuration file first
            if 'username' in self.configuration:
                username = self.configuration['username']
            else:
                username = input("Onelogin Username: ")
            self.username = username

    def load_password(self):
        """
        Load the password from keychain if we expect to be able to save the
        password in a keychain, or prompt the user for a password through
        stdin.
        """

        save_password = False
        reset_password = self.configuration.get('reset_password')

        # Do we have a password?
        if not self.has_password:
            # Can we load the password from the configuration?
            if self.configuration.get('password'):
                self.password = self.configuration['password']

            # Can we load the password from os keychain?
            if self.configuration.can_save_password:

                # Load the password from OS keychain
                self._load_password_from_keychain()

                # Could not find password in OS keychain
                if not self.has_password or reset_password:
                    # Ask user for password
                    self._prompt_user_password()
                    # Remember to save password
                    save_password = True

                if not self.has_password:
                    # We still don't have a password and have exhausted all
                    # places to load one from.
                    raise RuntimeError(
                        "Could not load password from secure store " +
                        "nor from user input"
                    )
            else:
                # Ask the user
                self._prompt_user_password()

            if save_password:
                # We decided to save the password
                print("Saving password to keychain...")
                self._save_password_to_keychain()

    def _prompt_user_password(self):
        self.password = getpass.getpass("Onelogin Password: ")

    def _load_password_from_keychain(self):
        self.password = keyring.get_password(self.SERVICE_NAME, self.username)

    def _save_password_to_keychain(self):
        keyring.set_password(self.SERVICE_NAME, self.username, self.password)



"""
Collections of entrypoints
"""
import sys
from os import environ

from onelogin_aws_login import OneloginAWS
from onelogin_aws_login.argparse import OneLoginAWSArgumentParser
from onelogin_aws_login.configuration import ConfigurationFile


def _load_config(parser, config_file: ConfigurationFile, args=sys.argv[1:]):
    cli_args = parser.parse_args(args)

    with open(cli_args.config_path, 'a+') as fp:
        fp.seek(0, 0)
        config_file.file = fp
        config_file.load()

        if (cli_args.configure or not config_file.is_initialised):
            config_file.initialise(cli_args.config_name)

    config_section = config_file.section(cli_args.config_name)

    if config_section is None or not config_section.has_required:
        sys.exit(
            "Configuration '{}' not defined. "
            "Please run 'onelogin-aws-login -c'".format(
                cli_args.config_name
            )
        )

    return config_section, cli_args


def login(args=sys.argv[1:]):
    """
    Entrypoint for `onelogin-aws-login`
    :param args:
    """

    debug = environ.get('ONELOGIN_AWS_CLI_DEBUG', '0') == '1'
    try:

        cfg = ConfigurationFile()
        parser = OneLoginAWSArgumentParser()
        config_section, args = _load_config(parser, cfg, args)

        config_section.set_overrides(vars(args))

        api = OneloginAWS(config_section)
        api.save_credentials()

    except Exception as e:
        if debug:
            raise e

        print(str(e))
        sys.exit(1)



import setuptools

PACKAGE_NAME = "onelogin_aws_login"

setuptools.setup(
    name=PACKAGE_NAME,
    version="1.1",
    author="",
    author_email="",
    description="",
    url="",
    packages=setuptools.find_packages(),
    include_package_data=True,
    python_requires=">=3.8",
    py_modules=[PACKAGE_NAME],
    install_requires=[
        "boto3",
        "onelogin",
        "keyring",
        "requests",
    ],
    entry_points={
        "console_scripts": [
            f"onelogin-aws-login = {PACKAGE_NAME}.cli:login"
        ]
    },
    zip_safe=False,
)



"""Static User Configuration models"""
import configparser

from onelogin_aws_login.userquery import user_choice


class ConfigurationFile(configparser.ConfigParser):
    """Represents a configuration ini file on disk"""

    DEFAULTS = dict(
        save_password=False,
        reset_password=False,
        duration_seconds=3600,
        auto_determine_ip_address=False,
        region=None
    )

    REQUIRED = [
        'base_uri',
        'client_id',
        'client_secret',
        'aws_app_id',
        'subdomain',
    ]

    def __init__(self, config_file=None):
        super().__init__(
            default_section='defaults',
        )

        self.file = config_file

        if self.file is not None:
            self.load()

    @property
    def has_defaults(self) -> bool:
        """True if the default section is non-empty"""
        return len(self.items(self.default_section)) > 0

    @property
    def is_initialised(self) -> bool:
        """True if there is at least one section"""
        return (len(self.sections()) > 0) or self.has_defaults

    def load(self):
        self.read_file(self.file)

    def initialise(self, config_name='defaults'):
        """
        Prompt the user for configurations, and save them to the
        onelogin-aws-cli config file
        """
        print("Configure Onelogin and AWS\n\n")
        config_section = self.section(config_name)
        if config_section is None:
            self.add_section(config_name)
            config_section = self.section(config_name)

        config_section['base_uri'] = user_choice(
            "Pick a Onelogin API server:", [
                "https://api.us.onelogin.com/",
                "https://api.eu.onelogin.com/"
            ]
        )

        print("\nOnelogin API credentials. These can be found at:\n"
              "https://admin.us.onelogin.com/api_credentials")
        config_section['client_id'] = input("Onelogin API Client ID: ")
        config_section['client_secret'] = input("Onelogin API Client Secret: ")
        print("\nOnelogin AWS App ID. This can be found at:\n"
              "https://admin.us.onelogin.com/apps")
        config_section['aws_app_id'] = input("Onelogin App ID for AWS: ")
        print("\nOnelogin subdomain is 'company' for login domain of "
              "'company.onelogin.com'")
        config_section['subdomain'] = input("Onelogin subdomain: ")

        self.save()

    def save(self):
        """Save this config to disk"""
        self.write(self.file)
        print("Configuration written to '{}'".format(
            self.file.name if hasattr(self.file, 'name') else self.file,
        ))

    def section(self, section_name: str):
        """
        Return a Section object representing a single section within the
        onelogin config file.

        :param section_name: Name of the section in the config file
        :return:
        """
        section_missing = not self.has_section(section_name)
        not_default = self.default_section != section_name
        if section_missing and not_default:
            return None
        return Section(section_name, self)


class Section(object):
    """Represents a single section in an ini file"""

    def __init__(self, section_name: str, config: ConfigurationFile):
        self.config = config
        self.section_name = section_name
        self._overrides = {}

    @property
    def has_required(self) -> bool:
        """
        Returns true if the section (including the defaults fallback)
        contains all the required keys.
        """
        return all([
            self.__contains__(item) for item in ConfigurationFile.REQUIRED
        ])

    @property
    def can_save_password(self) -> bool:
        """
        If the user has specified that the password can be saved to the system
        keychain
        :return:
        """
        return self.config.getboolean(
            self.section_name, "save_password",
            fallback=self.config.DEFAULTS['save_password']
        )

    @property
    def auto_determine_ip_address(self):
        return self.config.getboolean(
            self.section_name,
            "auto_determine_ip_address",
        )

    def set_overrides(self, overrides: dict):
        """
        Specify a dictionary values which take precedence over the existing
        values, but will not overwrite them in the config file.
        :param overrides:
        """
        self._overrides = {k: v for k, v in overrides.items() if v is not None}

    def __setitem__(self, key, value):
        self.config.set(self.section_name, key, value)

    def __getitem__(self, item):
        if item in self._overrides:
            return self._overrides[item]

        if self.config.has_option(self.section_name, item):
            return self.config.get(self.section_name, item)

        return self.config.DEFAULTS[item]

    def __contains__(self, item):
        return (item in self._overrides) or \
            self.config.has_option(self.section_name, item) or \
            item in self.config.DEFAULTS

    def get(self, item, default=None):
        if self.__contains__(item):
            return self.__getitem__(item)
        return default



"""
Argument Parser
"""

import argparse
import os

import pkg_resources

from onelogin_aws_login import DEFAULT_CONFIG_PATH, OneloginAWS


class OneLoginAWSArgumentParser(argparse.ArgumentParser):
    """Argument Parser separated into daemon and cli tool"""

    def __init__(self):
        super().__init__(description='Login to AWS with OneLogin')

        self.add_argument(
            '-C', '--config-name',
            action=EnvDefault, required=False,
            dest='config_name', default='defaults',
            help='Switch configuration name within config file'
        )

        self.add_argument(
            '--profile',
            action=EnvDefault, required=False,
            help='Specify profile name of credential',
        )

        self.add_argument(
            '--region',
            action=EnvDefault, required=False,
            default=None,
            help='Specify default region for AWS profile being updated'
        )

        self.add_argument(
            '-u', '--username',
            action=EnvDefault, required=False,
            help='Specify OneLogin username'
        )

        self.add_argument(
            '-p', '--password',
            action=EnvDefault, required=False,
            help='Specify OneLogin password'
        )

        self.add_argument(
            '-d', '--duration-seconds', type=int,
            dest='duration_seconds',
            action=EnvDefault, required=False,
            help='Specify duration seconds which depend on IAM role session '
                 'duration: https://aws.amazon.com/about-aws/whats-new/2018'
                 '/03/longer-role-sessions/'
        )

        self.add_argument(
            '--reset-password', dest='reset_password', action='store_true',
            help='Prompt the user for password even if the password is '
                 'stored in the OS keychain.', default=False,
        )

        self.add_argument(
            '--config-path',
            action=EnvDefault, required=False,
            dest='config_path', default=DEFAULT_CONFIG_PATH,
            help='Specify the configuration file path'
        )

        self.add_argument(
            '-c', '--configure', dest='configure', action='store_true',
            help='Configure OneLogin and AWS settings', default=False
        )

        version = pkg_resources.get_distribution(__package__).version
        self.add_argument(
            '-v', '--version', action='version',
            version="%(prog)s " + version
        )


class EnvDefault(argparse.Action):
    """Allow argparse values to be pulled from environment variables"""

    def __init__(self, required=True, default=None, **kwargs):

        if 'dest' in kwargs:
            name = 'ONELOGIN_AWS_CLI_' + kwargs['dest'].upper()
            # Fall back to the explicit command line default.
            default = os.environ.get(name, default)
            if 'type' in kwargs and default is not None:
                default = kwargs['type'](default)

        super().__init__(default=default, required=required, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, values)
